#region usings
using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;

using Microsoft.Kinect.Tools;
#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "Player", Category = "Kinect2Studio", Help = "Wrapper for KStudioPlayback class", Tags = "kinect2, kinect studio", Author = "ravazquez")]
	#endregion PluginInfo
	public class Kinect2StudioPlayerNode : IPluginEvaluate
	{
		#region fields & pins
		[Input("Client", IsSingle = true)]
		public ISpread<KStudioClient> FInputClient;
		
		[Input("EnableColor", IsSingle = true)]
		public ISpread<bool> FInputEnableColor;
		
		[Input("EnableDepth", IsSingle = true)]
		public ISpread<bool> FInputEnableDepth;
		
		[Input("EnableIR", IsSingle = true)]
		public ISpread<bool> FInputEnableIR;
		
		[Input("FilePath", IsSingle = true, StringType = StringType.Filename)]
		public ISpread<string> FInputFilePath;
		
		[Input("LoopCount", IsSingle = true)]
		public ISpread<uint> FInputLoopCount;
		
		[Input("Play", IsSingle = true)]
		public ISpread<bool> FInputPlay;
		
		//[Input("Pause", IsSingle = true)]
		//public ISpread<bool> FInputPause;
		
		//[Input("StepOnce", IsSingle = true)]
		//public ISpread<bool> FInputStepOnce;
		
		[Input("Stop", IsSingle = true)]
		public ISpread<bool> FInputStop;
		
		[Input("Enabled", IsSingle = true)]
		public ISpread<bool> FInputEnabled;

		[Output("Duration", IsSingle = true)]
		public ISpread<double> FOutputDuration;
		
		[Output("LoopIteration", IsSingle = true)]
		public ISpread<uint> FOutputLoopIteration;
		
		[Output("RelativeTime", IsSingle = true)]
		public ISpread<double> FOutputRelativeTime;
		
		[Output("Status", IsSingle = true)]
		public ISpread<string> FOutputStatus;
		
		private KStudioClient client;
		private KStudioPlayback playback;
		private uint loopCount = 0;
		private string filePath = ".";
		private string status = "Idle";
		#endregion fields & pins

		private void StartPlayback(string filePath, uint loopCount)
		{
			if (FInputClient.SliceCount > 0)
			{
				client = FInputClient[0];
			}
			if (client == null)
            {
                status = "Uninitialized client";
            }
			else
			{
				if (!client.IsServiceConnected)
				{
					client.ConnectToService();
				}
				if (playback != null && playback.State == KStudioPlaybackState.Paused)
				{
					playback.Resume();
					status = "Playing";
				}
				else{
					KStudioEventStreamSelectorCollection streamCollection = new KStudioEventStreamSelectorCollection();
					bool validated = true;
					List<Guid> streamGuids = InitializeStreamGuids();
					//check if selected streams are recordable
					foreach (Guid stream in streamGuids)
		            {
		                KStudioEventStream eventStream = client.GetEventStream(stream, KStudioEventStreamSemanticIds.KinectDefaultSensorProducer);
		                if (eventStream != null && !eventStream.IsPlaybackable)
		                {
		                    status = "Non-playbackable stream: " + eventStream.DataTypeName;
		                	validated = false;
		                }
		
		                streamCollection.Add(stream);
		            }
					//check dependencies between selected streams
					validated = validated || ValidateStreams(streamGuids);
					
					if (validated){
						KStudioPlaybackFlags flags = KStudioPlaybackFlags.IgnoreOptionalStreams;
						playback = client.CreatePlayback(filePath, streamCollection, flags);
						status = "Playing";
						playback.LoopCount = loopCount;
						playback.Start();	
					}
				}
			}
		}
		
		private void PausePlayback()
		{
			if(client != null)
			{
				if (playback != null)
				{
					if (playback.State == KStudioPlaybackState.Playing)
					{
						playback.Pause();
						status = "Paused";
					}
				}
			}
		}
		
		private void StepOnce()
		{
			if(client != null)
			{
				if (playback != null)
				{
					if (playback.State == KStudioPlaybackState.Paused)
					{
						playback.StepOnce();
						status = "Paused";
					}
				}
			}
		}
		
		private void StopPlayback()
		{
			if(client != null)
			{
				if (playback != null)
				{
					if (playback.State == KStudioPlaybackState.Playing || 
						playback.State == KStudioPlaybackState.Paused)
					{
						playback.Stop();
						client.DisconnectFromService();
						client = null;
						FOutputLoopIteration[0] = 0;
						FOutputRelativeTime[0] = 0;
						status = "Idle";
					}
				}
			}
		}
		
		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			if (FInputEnabled.SliceCount > 0 && FInputEnabled[0]){
				if (status == "Disabled"){
					status = "Idle";
				}
				//handle Play Input pin
				if (FInputPlay.SliceCount > 0 && FInputPlay[0])
				{
					if (FInputFilePath.SliceCount > 0)
						filePath = FInputFilePath[0];
					
					if (FInputLoopCount.SliceCount > 0)
						loopCount = FInputLoopCount[0];
					
					StartPlayback(filePath, loopCount);
				}
				/*
					This section was commented out because Pause causes the whole vvvv UI to freeze as long as the playback is Paused
				//handle Pause input pin
				if(FInputPause.SliceCount > 0 && FInputPause[0]){
					PausePlayback();
				}
				//handle StepOnce input pin
				if(FInputStepOnce.SliceCount > 0 && FInputStepOnce[0]){
					StepOnce();
				}
				*/
				//handle Stop input pin
				if (FInputStop.SliceCount > 0 && FInputStop[0])
				{
					StopPlayback();
				}
				//update output values
				if(playback != null)
				{
					if (client != null && client.IsServiceConnected)
					{
						FOutputDuration[0] = playback.Duration.TotalSeconds;
						FOutputLoopIteration[0] = playback.CurrentLoopIteration;
						FOutputRelativeTime[0] = playback.CurrentRelativeTime.TotalSeconds;
					}
					
					if (playback.State == KStudioPlaybackState.Stopped)
					{
						status = "Idle";
					}
				}
			}
			else{
				status = "Disabled";
			}
			
			FOutputStatus[0] = status;
		}
		
		private List<Guid> InitializeStreamGuids()
		{
			List<Guid> streamGuids = new List<Guid>();
			
			if(FInputEnableIR.SliceCount > 0 && FInputEnableIR[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.Ir);
			if(FInputEnableDepth.SliceCount > 0 && FInputEnableDepth[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.Depth);
			if(FInputEnableColor.SliceCount > 0 && FInputEnableColor[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.UncompressedColor);
			
			
			return streamGuids;
		}
	
        private bool ValidateStreams(IEnumerable<Guid> streamGuids)
        {
        	bool result = true;
            if (streamGuids == null)
        	{
                status = "No streams enabled";
        		result = false;
        	}
            if (streamGuids.Contains(KStudioEventStreamDataTypeIds.Ir) && !streamGuids.Contains(KStudioEventStreamDataTypeIds.Depth))
        	{
                status = "IR stream depends on Depth stream";
        		result = false;
        	}

            if (streamGuids.Contains(KStudioEventStreamDataTypeIds.Depth) && !streamGuids.Contains(KStudioEventStreamDataTypeIds.Ir))
        	{
                status = "Depth stream depends on IR stream";
        		result = false;
        	}

            //ToDo:  Check if Color and IR depend on each other
        	return result;
        }		
	}
}
