#region usings
using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;

using Microsoft.Kinect.Tools;
#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "Recorder", Category = "Kinect2Studio", Help = "Wrapper for KStudioRecording class", Tags = "")]
	#endregion PluginInfo
	public class Kinect2StudioRecorderNode : IPluginEvaluate
	{
		#region fields & pins
		[Input("Client", IsSingle = true)]
		public ISpread<KStudioClient> FInputClient;
		
		[Input("EnableBody", IsSingle = true)]
		public ISpread<bool> FInputEnableBody;
		
		[Input("EnableBodyIndex", IsSingle = true)]
		public ISpread<bool> FInputEnableBodyIndex;
		
		[Input("EnableColor", IsSingle = true)]
		public ISpread<bool> FInputEnableColor;
		
		[Input("EnableDepth", IsSingle = true)]
		public ISpread<bool> FInputEnableDepth;
		
		[Input("EnableIR", IsSingle = true)]
		public ISpread<bool> FInputEnableIR;
		
		[Input("FilePath", IsSingle = true, StringType = StringType.Directory)]
		public ISpread<string> FInputFilePath;
		
		[Input("BufferSizeInMegaBytes", IsSingle = true)]
		public ISpread<uint> FInputBufferSize;
		
		[Input("Record", IsSingle = true)]
		public ISpread<bool> FInputRecord;
		
		[Input("Stop", IsSingle = true)]
		public ISpread<bool> FInputStop;
		
		[Input("Enabled", IsSingle = true)]
		public ISpread<bool> FInputEnabled;

		[Output("Duration", IsSingle = true)]
		public ISpread<double> FOutputDuration;
		
		[Output("BufferUsage", IsSingle = true)]
		public ISpread<uint> FOutputBufferUsage;
		
		[Output("Status", IsSingle = true)]
		public ISpread<string> FOutputStatus;
		
		private KStudioClient client;
		private KStudioRecording recording;
		private string filePath = ".";
		private uint bufferSize = 1024;
		private string status = "Idle";
		#endregion fields & pins

		private void StartRecording(string filePath, uint bufferSize)
		{
			if (FInputClient.SliceCount > 0)
			{
				client = FInputClient[0];
			}
			if (client == null)
            {
                status = "Uninitialized client";
            }
			else
			{
				if (!client.IsServiceConnected)
				{
					client.ConnectToService();
				}
				KStudioEventStreamSelectorCollection streamCollection = new KStudioEventStreamSelectorCollection();
				bool validated = true;
				List<Guid> streamGuids = InitializeStreamGuids();
				//check if selected streams are recordable
				foreach (Guid stream in streamGuids)
	            {
	                KStudioEventStream eventStream = client.GetEventStream(stream, KStudioEventStreamSemanticIds.KinectDefaultSensorProducer);
	                if (eventStream != null && !eventStream.IsRecordable)
	                {
	                    status = "Non-recordable stream: " + eventStream.Name;
	                	validated = false;
	                }
	
	                streamCollection.Add(stream);
	            }
				//check dependencies between selected streams
				validated = validated || ValidateStreams(streamGuids);
				
				if (validated){
					KStudioRecordingFlags flags = KStudioRecordingFlags.GenerateFileName | KStudioRecordingFlags.IgnoreOptionalStreams;
					recording = client.CreateRecording(filePath, streamCollection, bufferSize, flags);
					status = "Recording";
					recording.Start();	
				}
			}
		}
		
		private void StopRecording()
		{
			if(client != null)
			{
				if (recording != null)
				{
					if (recording.State == KStudioRecordingState.Recording)
					{
						recording.Stop();
						FOutputBufferUsage[0] = 0;
						status = "Idle";
						client.DisconnectFromService();
						client = null;
					}
				}
			}
		}
		
		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			if (FInputEnabled.SliceCount > 0 && FInputEnabled[0]){
				if (status == "Disabled"){
					status = "Idle";
				}
				if (FInputRecord.SliceCount > 0 && FInputRecord[0])
				{
					if (FInputFilePath.SliceCount > 0)
						filePath = FInputFilePath[0];
					if (FInputBufferSize.SliceCount > 0)
						bufferSize = FInputBufferSize[0];
					
					StartRecording(filePath,bufferSize);
				}
				else if (FInputStop.SliceCount > 0 && FInputStop[0])
				{
					StopRecording();
				}
				if(recording != null && recording.State == KStudioRecordingState.Recording)
				{
					FOutputBufferUsage[0] = recording.BufferInUseSizeMegabytes;
					FOutputDuration[0] = recording.Duration.TotalSeconds;
				}
			}
			else{
				status = "Disabled";
			}
			
			FOutputStatus[0] = status;
		}
		
		private List<Guid> InitializeStreamGuids()
		{
			List<Guid> streamGuids = new List<Guid>();
			
			if(FInputEnableIR.SliceCount > 0 && FInputEnableIR[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.Ir);
			if(FInputEnableDepth.SliceCount > 0 && FInputEnableDepth[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.Depth);
			if(FInputEnableBody.SliceCount > 0 && FInputEnableBody[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.Body);
			if(FInputEnableBodyIndex.SliceCount > 0 && FInputEnableBodyIndex[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.BodyIndex);
			if(FInputEnableColor.SliceCount > 0 && FInputEnableColor[0])
				streamGuids.Add(KStudioEventStreamDataTypeIds.UncompressedColor);
			
			
			return streamGuids;
		}
	
        private bool ValidateStreams(IEnumerable<Guid> streamGuids)
        {
        	bool result = true;
            if (streamGuids == null)
        	{
                status = "No streams enabled";
        		result = false;
        	}
            if (streamGuids.Contains(KStudioEventStreamDataTypeIds.Ir) && !streamGuids.Contains(KStudioEventStreamDataTypeIds.Depth))
        	{
                status = "IR stream depends on Depth stream";
        		result = false;
        	}

            if (streamGuids.Contains(KStudioEventStreamDataTypeIds.Depth) && !streamGuids.Contains(KStudioEventStreamDataTypeIds.Ir))
        	{
                status = "Depth stream depends on IR stream";
        		result = false;
        	}

            //ToDo:  Check if Color and IR depend on each other
        	return result;
        }		
	}
}
